@@ -14,5 +14,12 @@ func init() {
 
 func registerSSLOptions(self *ToolOptions) error {
 	_, err := self.parser.AddGroup("ssl options", "", self.SSL)
-	return err
+	if err != nil {
+		return err
+	}
+	if self.enabledOptions.URI {
+		self.URI.AddKnownURIParameters(KnownURIOptionsSSL)
+	}
+	BuiltWithSSL = true
+	return nil
 }
